# $Id: ParserSkel.py,v 1.2 1999/10/07 08:26:09 leonard Exp $
# Copyright David Leonard, 1999. 

'''
	This is a parser driver that can be used to implement a parser
	based on the lists and dictionaries emitted by lalr_core.Grammar 
	object's tables() method. e.g.:

		import lalr_core
		p = lalr_core.new_parser()
		... #-- add productions, terminals etc
		t = p.tables()
		result = ParserSkel.parser(nexttoken, t)

	It is equivalent to calling the p.parse() method directly.
'''

class ParseError:
	def __init__(self, symbol, value):
		self.symbol = symbol
		self.value = value
	def __str__(self): return "ParseError before token %s (%s)" % (
		repr(self.symbol), repr(self.value))

def default_redfn(*args):
	if len(args) == 1:
		return args[0]
	return args

def parse(nexttok, tables, ptracef = None):
	(eof, goal, init, states) = tables

	inttype = type(0)

	#-- first element is simply a placeholder for the initial state
	sym_stack = [None]
	state_stack = [init]
	value_stack = [None]

	#-- avoid re-shifting eof symbol
	ateof = 0
	
	#-- load first token
	(next, next_value) = nexttok()
	if next is eof:
		ateof = 1

	while 1:
		if ptracef:
			#-- print trace of parse
			for i in range(1, len(sym_stack)):
			    ptracef.write("%d:%s/%s " % (
				    state_stack[i], sym_stack[i], 
				    `value_stack[i]`))
			ptracef.write(" . %s/%s\n" % (next, `next_value`))

		#-- find out what we can do
		actdict = states[state_stack[-1]]

		if not actdict.has_key(next):
			#-- we can't shift or reduce that symbol
			raise ParseError(next, next_value)

		action = actdict[next]

		if type(action) == inttype:
			#-- we can shift
			state_stack.append(action)
			sym_stack.append(next)
			value_stack.append(next_value)
			if not ateof:
				(next, next_value) = nexttok()
				if next is eof:
					ateof = 1
		else:
			#-- we can reduce
			nrhs, lhs, redfn = action
			if redfn is None:
				redfn = default_redfn
			value = apply(redfn, tuple(value_stack[-nrhs:]))

			actdict = states[state_stack[-nrhs-1]]
			if len(state_stack) == nrhs + 1 and \
			   not actdict.has_key(lhs):
				return value
			if not actdict.has_key(lhs):
				raise ParseError
			state_stack[-nrhs:] = [actdict[lhs]]
			value_stack[-nrhs:] = [value]
			sym_stack[-nrhs:] = [lhs]
