# $Id: regex.py,v 1.3 1999/10/07 08:26:08 leonard Exp $

#
# Another example, this time for parsing regular expressions.
#

import sys, traceback

from lalr_core import *
from ParserSkel import parse

g = new_parser()

BAR = '|';	g.add_terminal(BAR)
STAR = '*';	g.add_terminal(STAR)
LEFT = '(';	g.add_terminal(LEFT)
RIGHT = ')';	g.add_terminal(RIGHT)
CC = 'cc';	g.add_terminal(CC)

RegExpn = "RegExpn";	g.add_nonterminal(RegExpn)
Term = "Term";		g.add_nonterminal(Term)
Primary = "Primary";	g.add_nonterminal(Primary)
Factor = "Factor";	g.add_nonterminal(Factor)

class Atom:
	def __init__(self, a):
		self.a = a
	def __repr__(self): return str(self.a)
class Concat:
	def __init__(self, a, b):
		self.a = a
		self.b = b
	def __repr__(self): return "("+repr(self.a)+repr(self.b)+")"
class Alternate:
	def __init__(self, a, b):
		self.a = a
		self.b = b
	def __repr__(self): return "("+repr(self.a)+"|"+repr(self.b)+")"
class Kleene:
	def __init__(self, a):
		self.a = a
	def __repr__(self): return "("+repr(self.a)+")*"

g.add_production(RegExpn,	[RegExpn, BAR, Term],
					lambda r,_,t: Alternate(r,t))
g.add_production(RegExpn,	[Term])
g.add_production(Term,		[Term, Primary],
					lambda t,p:   Concat(t,p))
g.add_production(Term,		[Primary])
g.add_production(Primary,	[Factor, STAR],
					lambda f,_:   Kleene(f))
g.add_production(Primary,	[Factor])
g.add_production(Factor,	[LEFT, RegExpn, RIGHT],
					lambda _,e,_: e)
g.add_production(Factor,	[CC],
					lambda cc:    Atom(cc))

# simple goal production... this should be done automatically??? all that
# should be needed is the eof as the goal nonterminal should be implicit
G = "G";	g.add_nonterminal(G)
eof = 'eof';	g.add_terminal(eof)
g.add_production(G, [RegExpn, eof], lambda a,_: a)
g.set_goal_eof(G, eof)

# lexer
lexermap = {
	'*':  STAR,
	'|':  BAR,
	'(':  LEFT,
	')':  RIGHT
}
nextchar = None
def initnexttoken():
	global nextchar
	nextchar = None
def getchar():
	global nextchar
	nxt = nextchar
	nextchar = sys.stdin.read(1)
	return nxt
def nexttoken():
    #print "nexttoken:",
    global nextchar, lexermap
    if nextchar is None: 
	getchar()	# (re-)prime getchar()
    while 1:
	if nextchar == '': 
		print "(eof)"
		raise SystemExit
	if nextchar in ('\n', '\r'):
		nextchar = None
		ret = eof,None
		break
	if lexermap.has_key(nextchar):	
		ret = lexermap[getchar()],None
		break
	#if nextchar == '\\':
	#	getchar()
	#	ret = CC,getchar()
	if nextchar in (' ', '\t'):
		getchar()
		continue
	ret = CC, getchar()
	break
    #print "->",`ret`
    return ret


#-- we use the ParserSkel.parser() driver instead of the lalr_core.parser().
#   The latter is written in C and the former in python. Both should take
#   about the same amount of time.

tables = g.tables()
# g.set_ptrace(sys.stderr)
while 1:
    try:
	print "Enter expression: ",
	initnexttoken()
	#res = g.parse(nexttoken)				# C
	res = parse(nexttoken, tables, ptracef = sys.stderr)	# Python
	print "> ", res
    except ParseError:
	traceback.print_exc()

